module interpolation

!>>>>>
!>
!>     INTERPOLATION ROUTINE
!>     USED WITH RHO, GRAD{RHO} AND HESS{RHO}
!>
!>>>>>

use kind_numbers
use constants_module
use grid_module
use rho_module
implicit none
private

public :: interp
intrinsic :: modulo,present,real
intrinsic :: int

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
subroutine interp(r,dens,grad,tot_val,hess)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!  interpolates rho given on a regular grid defined in "grid_module"
!  a(4,3),ap(4,3),app(4,3), are coefficient for the interpolation
!  s(3) and ss(3) are fractions of the incrementation in the interval ] 0,1[
!  the interpolation is based on a tricubic Lagrange method;
!  if x C [i,i+1] then the terms included are
!  i-1,i,i+1,i+2 corresponding to 1,2,3,4 in the table "a"
!
real(kind=double), intent(out) :: dens       !!! output density
real(kind=double), dimension(3), intent(in) :: r     !!! input position
real(kind=double), dimension(3), intent(out) :: grad !!! output gradient
real(kind=double), dimension(3,3), intent(out), optional :: hess !!! output hessian matrix
logical, intent(in) :: tot_val
!
integer(kind=int32) :: l,m,n,ll,mm,nn
integer(kind=int32), dimension(3) :: is
real(kind=double) :: tmp
real(kind=double), dimension(3) :: ss
real(kind=double), dimension(4,3) :: a,ap,app
logical :: with_hess
!
!  initialization of data
!
if(present(hess)) then
  with_hess=.true.
else
  with_hess=.false.
end if
dens=zero
grad(:)=zero
if(with_hess) hess(:,:)=zero
do l=1,3
  tmp=r(l)
  is(l)=int(tmp)+1
  ss(l)=tmp+one-real(is(l),double)
end do
!
!  if new position is out of box, shift of i-1, i, i+1 and i+2
!
do l=1,3
  if(ss(l) < zero) then
    is(l)=is(l)-1
    ss(l)=ss(l)+one
  end if
  if(ss(l) >= one) then
    is(l)=is(l)+1
    ss(l)=ss(l)-one
  end if
end do
!
!  coefficient for the interpolation
!
do l=1,3
  tmp=ss(l)
  a(1,l)=((-tmp*over6+half)*tmp-over3)*tmp
  a(2,l)=((tmp*half-one)*tmp-half)*tmp+one
  a(3,l)=((-tmp*half+half)*tmp+one)*tmp
  a(4,l)=(tmp*tmp-one)*tmp*over6
!
  ap(1,l)=(-tmp*half+one)*tmp-over3
  ap(2,l)=(tmp*t3over2-two)*tmp-half
  ap(3,l)=(-tmp*t3over2+one)*tmp+one
  ap(4,l)=tmp*tmp*half-over6
!
  if(with_hess) then
    app(1,l)=-tmp+one
    app(2,l)=tmp*three-two
    app(3,l)=-tmp*three+one
    app(4,l)=tmp
  end if
end do
!
!  interpolation of density, gradient and hessian
!
do n=1,4
  nn=(n-2)+is(3)
  nn=modulo(nn,np3)
  if(nn <= 0) nn=nn+np3
  do m=1,4
    mm=(m-2)+is(2)
    mm=modulo(mm,np2)
    if(mm <= 0) mm=mm+np2
    do l=1,4
      ll=(l-2)+is(1)
      ll=modulo(ll,np1)
      if(ll <= 0) ll=ll+np1
      tmp=rho(ll,mm,nn)
! interpolation of "total" or "valence" density according to "tot_val" flag
      if(tot_val) then
        dens=dens+tmp*a(l,1)*a(m,2)*a(n,3)
      else
        dens=dens+rhov(ll,mm,nn)*a(l,1)*a(m,2)*a(n,3)
      end if
! computes gradient in "grid" coordinates
      grad(1)=grad(1)+tmp*ap(l,1)*a(m,2)*a(n,3)
      grad(2)=grad(2)+tmp*a(l,1)*ap(m,2)*a(n,3)
      grad(3)=grad(3)+tmp*a(l,1)*a(m,2)*ap(n,3)
! computes hessian in "grid" coordinates
      if(with_hess) then
        hess(1,1)=hess(1,1)+tmp*app(l,1)*a(m,2)*a(n,3)
        hess(1,2)=hess(1,2)+tmp*ap(l,1)*ap(m,2)*a(n,3)
        hess(1,3)=hess(1,3)+tmp*ap(l,1)*a(m,2)*ap(n,3)
        hess(2,2)=hess(2,2)+tmp*a(l,1)*app(m,2)*a(n,3)
        hess(2,3)=hess(2,3)+tmp*a(l,1)*ap(m,2)*ap(n,3)
        hess(3,3)=hess(3,3)+tmp*a(l,1)*a(m,2)*app(n,3)
      end if
    end do
  end do
end do
if(with_hess) then
  hess(2,1)=hess(1,2)
  hess(3,1)=hess(1,3)
  hess(3,2)=hess(2,3)
end if
!
end subroutine interp

end module interpolation
