module math_module

use kind_numbers
use constants_module
implicit none
private

public :: matinv,eigen
intrinsic :: abs,reshape,sign,sqrt

real(kind=double), public :: EpsOff=1.0e-10_double
real(kind=double), dimension(3,3), parameter, private :: IdentityMatrix=reshape( &
(/ one,zero,zero, zero,one,zero, zero,zero,one /),(/3,3/))

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
subroutine matinv(mat,invmat)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!   Inversion of a symmetric (3x3) matrix using GAUSS-JORDAN elimination
!
real(kind=double), dimension(3,3), intent(in) :: mat
real(kind=double), dimension(3,3), intent(out) :: invmat
!
integer(kind=int32) :: i,j,k
integer(kind=int32), dimension(3) :: ik,jk
real(kind=double) :: t,matmax
real(kind=double), dimension(3) :: coef
!
invmat=mat
do i=1,3
  t=sqrt(abs(invmat(i,i)))
  if(t == zero) then
    return
  else
    coef(i)=one/t
  end if
end do
!
!  Scale matrix invmat
!
do i=1,3
  do j=1,3
    invmat(i,j)=invmat(i,j)*coef(i)*coef(j)
  end do
end do
!
!  invert invmat
!
do k=1,3
!
!        matrix ordening by largest values
!
  matmax=zero
  do i=k,3
    do j=i,3
      if(abs(invmat(i,j)) > abs(matmax)) then
        matmax=invmat(i,j)
        ik(k)=i
        jk(k)=j
      end if
    end do
  end do
  if(matmax == zero) then
    return
  end if
!
!        interchange row and columns to put matmax in a(k,k)
!
  i=ik(k)
  if(i > k) then
    do j=1,3
      t=invmat(k,j)
      invmat(k,j)=invmat(i,j)
      invmat(i,j)=-t
    end do
  end if
  j=jk(k)
  if(j > k) then
    do i=1,3
      t=invmat(i,k)
      invmat(i,k)=invmat(i,j)
      invmat(i,j)=-t
    end do
  end if
!
!        accumulate elements of inv matrix
!
  do i=1,3
    if(i /= k) invmat(i,k)=-invmat(i,k)/matmax
  end do
  do i=1,3
    do j=1,3
      if(i /= k.and.j /= k) invmat(i,j)=invmat(i,j)+invmat(i,k)*invmat(k,j)
    end do
  end do
  do j=1,3
    if(j /= k) invmat(k,j)=invmat(k,j)/matmax
  end do
  invmat(k,k)=one/matmax
end do
!
!     restore ordering of matrix
!
do k=3,1,-1
  j=ik(k)
  if(j > k) then
    do i=1,3
      t=invmat(i,k)
      invmat(i,k)=-invmat(i,j)
      invmat(i,j)=t
    end do
  end if
  i=jk(k)
  if(i > k) then
    do j=1,3
      t=invmat(k,j)
      invmat(k,j)=-invmat(i,j)
      invmat(i,j)=t
    end do
  end if
end do
!
!     rescale invmat
!
do i=1,3
  do j=1,3
    invmat(i,j)=invmat(i,j)*coef(i)*coef(j)
  end do
end do
!
end subroutine matinv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
subroutine eigen(SymMatrix,eigenvalues,eigenvectors)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!    Determines eigenvalues and eigenvectors of a symmetrical
!    (3x3) matrix, sort by decreasing eigenvalues
!
real(kind=double), dimension(3), intent(out) :: eigenvalues
real(kind=double), dimension(3,3), intent(out) :: eigenvectors
real(kind=double), dimension(3,3), intent(in) :: SymMatrix !! real  symmetric matrix
!
integer(kind=int32) :: i,j,k
real(kind=double) :: ResOff,TestOff
real(kind=double) :: u,v,w,s,c,c2,s2,cs
real(kind=double) :: bki,bkj,bii,bjj,bij
real(kind=double), dimension(3,3) :: Matrix    !! dummy symmetric matrix
!
Matrix=SymMatrix
eigenvectors=IdentityMatrix
ResOff=zero
!
do i=1,2
  do j=i+1,3
    ResOff=ResOff+Matrix(i,j)**2
  end do
end do
!
ResOff=sqrt(two*ResOff)
!
if(ResOff >= EpsOff) then
  TestOff=ResOff*EpsOff*over3
!
  do
    if(ResOff < TestOff) exit
    ResOff=ResOff*over3
    do i=1,2
      do j=i+1,3
        if(abs(Matrix(i,j)) > ResOff) then
          u=-Matrix(i,j)
          v=(Matrix(i,i)-Matrix(j,j))*half
          w=sign(one,v)*u/sqrt(u**2+v**2)
          s=w/sqrt(two*(one+sqrt(one-w**2)))
          c=sqrt(one-s**2)
          c2=c**2
          s2=s**2
          cs=c*s
!
          do k=1,3
            if(k /= i.and.k /= j) then
              bki=Matrix(k,i)*c-Matrix(k,j)*s
              bkj=Matrix(k,i)*s+Matrix(k,j)*c
              Matrix(k,i)=bki
              Matrix(k,j)=bkj
              Matrix(i,k)=Matrix(k,i)
              Matrix(j,k)=Matrix(k,j)
            end if
            bki=eigenvectors(k,i)*c-eigenvectors(k,j)*s
            bkj=eigenvectors(k,i)*s+eigenvectors(k,j)*c
            eigenvectors(k,i)=bki
            eigenvectors(k,j)=bkj
          end do
!
          bii=Matrix(i,i)*c2+Matrix(j,j)*s2-two*Matrix(i,j)*cs
          bjj=Matrix(i,i)*s2+Matrix(j,j)*c2+two*Matrix(i,j)*cs
          bij=(Matrix(i,i)-Matrix(j,j))*cs+Matrix(i,j)*(c2-s2)
          Matrix(i,i)=bii
          Matrix(j,j)=bjj
          Matrix(i,j)=bij
          Matrix(j,i)=bij
!
        end if
      end do
    end do
  end do
end if
!
do i=1,2
  do j=i+1,3
    if(Matrix(i,i) < Matrix(j,j)) then
      u=Matrix(i,i)
      Matrix(i,i)=Matrix(j,j)
      Matrix(j,j)=u
      do k=1,3
        u=eigenvectors(k,i)
        eigenvectors(k,i)=eigenvectors(k,j)
        eigenvectors(k,j)=u
      end do
    end if
  end do
end do
do i=1,3
  eigenvalues(i)=Matrix(i,i)
end do
!
end subroutine eigen

end module math_module
